/* $Id: svgalib.c,v 1.15 1998/12/04 21:34:25 becka Exp $
******************************************************************************

   SVGAlib wrapper for LibGGI - the vga_* functions

   Copyright (C) 1997 Jason McMullan	[jmcc@ggi-project.org]
   Copyright (C) 1998 Todd T. Fries	[toddf@acm.org]
   Copyright (C) 1998 Marcus Sundberg	[marcus@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include "svgaggi_internal.h"

static struct modeinfo_s {
	int width;
	int height;
	ggi_graphtype gt;
} ModeInfo[] = {
	{  GGI_AUTO, GGI_AUTO,GT_TEXT16}, /* #define TEXT     0 */
	{ 320, 200, GT_4BIT}, /* #define G320x200x16  1 */
	{ 640, 200, GT_4BIT}, /* #define G640x200x16  2 */
	{ 640, 350, GT_4BIT}, /* #define G640x350x16  3 */
	{ 640, 480, GT_4BIT}, /* #define G640x480x16  4 */

	{ 320, 200, GT_8BIT}, /* #define G320x200x256 5 */
	{ 320, 240, GT_8BIT}, /* #define G320x240x256 6 */
	{ 320, 400, GT_8BIT}, /* #define G320x400x256 7 */
	{ 360, 480, GT_8BIT}, /* #define G360x480x256 8 */

	{ 640, 480, GT_1BIT}, /* #define G640x480x2   9 */
  
	{ 640, 480, GT_8BIT}, /* #define G640x480x256 10 */
	{ 800, 600, GT_8BIT}, /* #define G800x600x256 11 */
	{1024, 768, GT_8BIT}, /* #define G1024x768x256 12 */
	{1280,1024, GT_8BIT}, /* #define G1280x1024x256 13 */

	{ 320, 200, GT_15BIT}, /* #define G320x200x32K 14 */
	{ 320, 200, GT_16BIT}, /* #define G320x200x64K 15 */
	{ 320, 200, GT_24BIT}, /* #define G320x200x16M 16 */

	{ 640, 480, GT_15BIT}, /* #define G640x480x32K 17 */
	{ 640, 480, GT_16BIT}, /* #define G640x480x64K 18 */
	{ 640, 480, GT_24BIT}, /* #define G640x480x16M 19 */
  
	{ 800, 600, GT_15BIT}, /* #define G800x600x32K 20 */
	{ 800, 600, GT_16BIT}, /* #define G800x600x64K 21 */
	{ 800, 600, GT_24BIT}, /* #define G800x600x16M 22 */

	{1024, 768, GT_15BIT}, /* #define G1024x768x32K 23 */
	{1024, 768, GT_16BIT}, /* #define G1024x768x64K 24 */
	{1024, 768, GT_24BIT}, /* #define G1024x768x16M 25 */

	{1280, 1024, GT_15BIT}, /* #define G1280x1024x32K 26 */
	{1280, 1024, GT_16BIT}, /* #define G1280x1024x64K 27 */
	{1280, 1024, GT_24BIT}, /* #define G1280x1024x16M 28 */

	{ 800, 600, GT_4BIT},  /* #define G800x600x16 29 */
	{1024, 768, GT_4BIT},  /* #define G1024x768x16 30 */
	{1280,1024, GT_4BIT},  /* #define G1280x1024x16 31 */

	{ 720, 348, GT_1BIT},  /* #define G720x348x2 32	 Hercules emulation mode */

	{ 320, 200, GT_32BIT}, /* #define G320x200x16M32 33 */
	{ 640, 480, GT_32BIT}, /* #define G640x480x16M32 34 */
	{ 800, 600, GT_32BIT}, /* #define G800x600x16M32 35 */
	{1024, 768, GT_32BIT}, /* #define G1024x768x16M32 36 */
	{1280, 1024, GT_32BIT},/* #define G1280x1024x16M32 37 */

	{1152, 864, GT_4BIT},  /* #define G1152x864x16 38 */
	{1152, 864, GT_8BIT},  /* #define G1152x864x256 39 */
	{1152, 864, GT_15BIT}, /* #define G1152x864x32K 40 */
	{1152, 864, GT_16BIT}, /* #define G1152x864x64K 41 */
	{1152, 864, GT_24BIT}, /* #define G1152x864x16M 42 */
	{1152, 864, GT_32BIT}, /* #define G1152x864x16M32 43 */

	{1600,1200, GT_4BIT},  /* #define G1600x1200x16 44 */
	{1600,1200, GT_8BIT},  /* #define G1600x1200x256 45 */
	{1600,1200, GT_15BIT}, /* #define G1600x1200x32K 46 */
	{1600,1200, GT_16BIT}, /* #define G1600x1200x64K 47 */
	{1600,1200, GT_24BIT}, /* #define G1600x1200x16M 48 */
	{1600,1200, GT_32BIT}, /* #define G1600x1200x16M32 49 */
};


int _ggigsw_async = 0;
int _ggigsw_but2key = 0;

static int _ggigsw_pflip = 0;
static int _ggigsw_curframe = 0;
static int _ggigsw_pageemu = 0;
static int _ggigsw_curpage = 0;
static unsigned char *_ggigsw_realfb = NULL;
static unsigned char *_ggigsw_realfb2 = NULL;
static unsigned char *_ggigsw_pagefb = NULL;

int _ggigsw_inggi = -2;   /* Current active entry of the ModeInfo table */
ggi_visual_t _ggigsw_visual;

static int                      _ggigsw_fbmem = 0;
static int                      _ggigsw_pages = 1;
static ggi_mode                 _ggigsw_mode;
static const ggi_directbuffer  *_ggigsw_dbuf;
static const ggi_directbuffer  *_ggigsw_dbuf2;
static int                      _ggigsw_stride;
static int                      _ggigsw_curmodenum;
static int                      _ggigsw_wantmouse = 0;

/* This is used by the svgalib target to avoid infinite recursion */
#define GSW_MAGIC	(-4711)
int __svgalib_tty_fd = GSW_MAGIC;

/**********************
 * Exported variables *
 **********************/

unsigned char *graph_mem = NULL;

/****************
 * Mode setting *
 ****************/
static void _vga_exit(int arg)
{
	ggiExit();
	exit(arg);
}

int vga_init(void)
{
	int i;
	char *str;

	if (ggiInit() < 0) {
		ggiPanic("SVGAlib wrapper: unable to init LibGGI\n");
	}

	if ((_ggigsw_visual = ggiOpen(NULL)) == NULL) {
		ggiPanic("SVGAlib wrapper: unable to open default visual\n");
	}

	if ((str = getenv("GSW_ASYNC"))) {
		/* We call ggiFlush() in mouse_update() or keyboard_update() */
		ggiSetFlags(_ggigsw_visual, GGIFLAG_ASYNC);
		if (strcmp(str, "mouse") == 0) {
			_ggigsw_async = GSW_ASYNCMOUSE;
			_GSWPRINT("Using async-mouse mode\n");
		} else {
			_ggigsw_async = GSW_ASYNCKEY;
			_GSWPRINT("Using async-key mode\n");
		}
	}
	if (getenv("GSW_PAGEEMU")) {
		/* Emulate a paged mode */
		_ggigsw_pageemu = GSW_WANTIT;
		_ggigsw_curpage = 0;
		if ((_ggigsw_pagefb = malloc(GSW_PAGESIZE)) == NULL) {
			ggiPanic("malloc() failed!\n");
		}
		_GSWPRINT("Using page-emulating mode\n");
	}
	if (getenv("GSW_DB")) {
		_ggigsw_pflip = GSW_WANTIT;
	}
	if (getenv("GSW_BUT2KEY")) {
		_ggigsw_but2key = GSW_HAVEIT;
	}
	for (i = 1; i < 31; i++)
		signal(i, _vga_exit);

	_ggigsw_inggi = -1;
	return 0;
}

int vga_setmode(int mode)
{
	int x, y, m, err = -1;
	char buff[1024];

	if (mode < 0 || mode > __GLASTMODE) 
		return -1;

	if (_ggigsw_inggi < -1) 
		vga_init();

	x = ModeInfo[mode].width;
	y = ModeInfo[mode].height;
	m = ModeInfo[mode].gt;
	_ggigsw_inggi = mode;

	if (m == GT_TEXT16) {
		ggiSetTextMode(_ggigsw_visual, x, y, GGI_AUTO, GGI_AUTO, 
			       GGI_AUTO, GGI_AUTO, GT_TEXT16);
		return 0;
	} else {
		ggi_mode gmod;
		gmod.visible.x = x;
		gmod.visible.y = y;
		gmod.virt.x = gmod.virt.y = GGI_AUTO;
		gmod.graphtype = m;
		gmod.dpp.x = gmod.dpp.y = GGI_AUTO;

		if (_ggigsw_pflip) {
			gmod.frames = 2;
			if ((err = ggiSetMode(_ggigsw_visual, &gmod))
			    != 0) {
				_ggigsw_pflip = GSW_WANTIT;
			} else {
				_ggigsw_pflip = GSW_HAVEIT;
			}				
		}
		
		if (err && ggiSetGraphMode(_ggigsw_visual, 
					   x, y, GGI_AUTO, GGI_AUTO, m)) {
			sprintf(buff, "Can't set graphics mode of %dx%d in %s colors\n",
				x, y,
				(m == GT_1BIT) ? "2" :
				(m == GT_4BIT) ? "16" :
				(m == GT_8BIT) ? "256" :
				(m == GT_15BIT) ? "32k" :
				(m == GT_16BIT) ? "64k" :
				(m == GT_24BIT) ? "16 million"  :
				(m == GT_32BIT) ? "16 million (32 bit)" :
				"an unknown number of");
			ggiPanic(buff);
		}
		graph_mem = _ggigsw_realfb = NULL;
	}
	ggiGetMode(_ggigsw_visual, &_ggigsw_mode);
	fprintf(stderr, "LibGGI SVGAlib-wrapper: ");
	ggiFPrintMode(stderr, &_ggigsw_mode);
	putc('\n', stderr);
	if ((_ggigsw_dbuf = ggiDBGetBuffer(_ggigsw_visual, 0)) != NULL) {
		if ((_ggigsw_dbuf->type & GGI_DB_SIMPLE_PLB)) {
			_ggigsw_stride = _ggigsw_dbuf->buffer.plb.stride;
			graph_mem = _ggigsw_realfb = _ggigsw_dbuf->write;
			_ggigsw_fbmem = _ggigsw_stride*y;
			_ggigsw_pages = (_ggigsw_fbmem+GSW_PAGESIZE-1) 
						/ GSW_PAGESIZE;
			if (_ggigsw_pageemu) {
				if (_ggigsw_fbmem > GSW_PAGESIZE) {
					graph_mem = _ggigsw_pagefb;
					_ggigsw_pageemu = GSW_HAVEIT;
				} else {
					_ggigsw_pageemu = GSW_WANTIT;
				}
			}
		} else {
			_ggigsw_dbuf = NULL;
		}
	}
	if (_ggigsw_pflip == GSW_HAVEIT) {
		if ((_ggigsw_dbuf2 = ggiDBGetBuffer(_ggigsw_visual, 1))
		    != NULL
		    && (_ggigsw_dbuf->type & GGI_DB_SIMPLE_PLB)
		    && _ggigsw_dbuf2->buffer.plb.stride
		    == _ggigsw_stride) {
			_ggigsw_realfb2 = _ggigsw_dbuf2->write;
		} else {
			_ggigsw_pflip = GSW_WANTIT;
		}
	}

	_ggigsw_curmodenum = mode;

	if (_ggigsw_wantmouse) {
		mouse_setxrange(0, x - 1);
		mouse_setyrange(0, y - 1);
		mouse_setwrap(MOUSE_NOWRAP);
	}

	return 0;
}


int vga_hasmode(int mode)
{
	int x, y, m;

	if (mode < 0 || mode > __GLASTMODE)
		return 0;

	if (_ggigsw_inggi < -1) 
		vga_init();

	x = ModeInfo[mode].width;
	y = ModeInfo[mode].height;
	m = ModeInfo[mode].gt;

	if (ggiCheckGraphMode(_ggigsw_visual, x, y, GGI_AUTO, GGI_AUTO, m, NULL)
	    == 0
	    || mode == 0) {
		return 1;
	}

	return 0;
}
  
int vga_setflipchar(int c)
{
	return 0;
}


int vga_clear(void)
{
	ggi_color col = {0, 0, 0};

	ggiSetGCForeground(_ggigsw_visual,
			   ggiMapColor(_ggigsw_visual, &col));
	ggiFillscreen(_ggigsw_visual);

	return 0;
}

int vga_flip(void)
{
	/* BUGBUG - should flip to the text VT, 
	 * or back again to the graphics VT
	 */
	return -1;
}


int vga_getxdim(void)
{
	if (_ggigsw_inggi < 0)
		return 0;

	return ModeInfo[_ggigsw_inggi].width;
}


int vga_getydim(void)
{
	if (_ggigsw_inggi < 0)
		return 0;

	return ModeInfo[_ggigsw_inggi].height;
}

int vga_getcolors(void)
{
	if (_ggigsw_inggi < 0)
		return 0;
   
	return (1 << GT_DEPTH(_ggigsw_mode.graphtype));
}

/*********************************
 *     P a l e t t e  I n f o    *
 *********************************/

int vga_setpalette(int index, int red, int green, int blue)
{
	ggi_color col;

	col.r = red   << COLORSHIFT;
	col.g = green << COLORSHIFT;
	col.b = blue  << COLORSHIFT;

	return (ggiSetPalette(_ggigsw_visual, index, 1, &col) ? -1 : 0);
}

int vga_getpalette(int index, int *red, int *green, int *blue)
{
	ggi_color col;

	if (ggiGetPalette(_ggigsw_visual, index, 1, &col))
		return -1;

	*red   = col.r >> COLORSHIFT;
	*green = col.g >> COLORSHIFT;
	*blue  = col.b >> COLORSHIFT;

	return 0;
}


int vga_setpalvec(int start, int num, int *pal)
{
	ggi_color cols[256];
	int i;

	for (i = start; i < num; i++) {
		cols[i].r = pal[0] << COLORSHIFT;
		cols[i].g = pal[1] << COLORSHIFT;
		cols[i].b = pal[2] << COLORSHIFT;
		pal += 3;
	}
	
	return (ggiSetPalette(_ggigsw_visual, start, num, cols) ? -1 : 0);
}

int vga_getpalvec(int start, int num, int *pal)
{
	ggi_color cols[256];
	int i;

	if (ggiGetPalette(_ggigsw_visual, start, num, cols)) {
		return -1;
	}

	for (i = start;i<num;i++) {
		pal[0] = cols[i].r >> COLORSHIFT;
		pal[1] = cols[i].g >> COLORSHIFT;
		pal[2] = cols[i].b >> COLORSHIFT;
		pal += 3;
	}
	
	return 0;
}


/******************
 * Screen control *
 ******************/

int vga_screenoff(void)
{
	/* BUGBUG - this should 'blank' the screen .. */
	return 0;
}
  
int vga_screenon(void)
{ 
	/* BUGBUG - this should 'unblank' the screen .. */
	return 0;
}


/********************
 * Drawing Commands *
 ********************/


int vga_setcolor(int color)
{
	return (ggiSetGCForeground(_ggigsw_visual, color) ? -1 : 0);
}

int vga_drawpixel(int x, int y)
{
	return (ggiDrawPixel(_ggigsw_visual, x, y) ? -1 : 0);
}

int vga_drawline(int x1, int y1, int x2, int y2)
{
	return (ggiDrawLine(_ggigsw_visual, x1, y1, x2, y2) ? -1 : 0);
}

int vga_drawscanline(int line, unsigned char *colors)
{
	int xlen;

	if (_ggigsw_inggi < 0)
		return -1;

	xlen = ModeInfo[_ggigsw_inggi].width;

	return (ggiPutHLine(_ggigsw_visual, 0, line, xlen, colors)
		? -1 : 0);
}

int vga_drawscansegment(unsigned char *colors, int x, int y, int length)
{
	return (ggiPutHLine(_ggigsw_visual, x, y, length, colors) ? -1 : 0);
}

int vga_getpixel(int x, int y)
{
	int pixel;

	return (ggiGetPixel(_ggigsw_visual, x, y, &pixel) ? pixel : -1);
}


int vga_getch(void)
{
	return ggiGetc(_ggigsw_visual);
}


int vga_dumpregs(void)
{
	/* BUGBUG - We should dump the mode-info here... */
	return 0;
}


/***************************
 * VGA Lib v1.2 extensions *
 ***************************/



vga_modeinfo *vga_getmodeinfo(int mode)
{
	static vga_modeinfo minfo;
	char *env;

	memset(&minfo, 0, sizeof(vga_modeinfo));

	if (mode < 0 || mode > __GLASTMODE)
		return NULL;
  
	minfo.width = ModeInfo[mode].width;
	minfo.height = ModeInfo[mode].height;
	minfo.bytesperpixel = GT_SIZE(ModeInfo[mode].gt)/8;
	minfo.colors = 1 << GT_DEPTH(ModeInfo[mode].gt);
	if ((env = getenv("GSW_WIDTH")) != NULL) {
		minfo.linewidth = atoi(env);
	} else {
		minfo.linewidth = minfo.width * minfo.bytesperpixel;
	}
	minfo.maxlogicalwidth = minfo.linewidth;
	minfo.maxpixels = 1600*1200; /* Just a random number */

	minfo.flags |= CAPABLE_LINEAR; /* We have no way of knowing without
					  actually setting the mode, so we set
					  the flag and hope that the app is
					  smart enough to check the return
					  value from vga_setlinearaddresing()
				       */
	if (mode == _ggigsw_curmodenum && _ggigsw_dbuf) {
		minfo.flags |= IS_LINEAR; /* A linear capable mode is always
					     linear when running on LibGGI
					  */
	}
	return &minfo;
}

int vga_getdefaultmode(void)
{
	return (_ggigsw_inggi < 0) ? G320x200x256 : _ggigsw_inggi;
}

int vga_getcurrentmode(void)
{
	return (_ggigsw_inggi < 0) ? -1 : _ggigsw_inggi;
}


int vga_getcurrentchipset(void)
{
	return UNDEFINED;
}

char *vga_getmodename(int mode)
{
	static char buff[1024];
	int m;

	m = ModeInfo[mode].gt;

	if (ModeInfo[mode].gt == GT_TEXT16)
		sprintf(buff, "TEXT16");
	else
		sprintf(buff, "G%dx%dx%s",
			ModeInfo[mode].width, ModeInfo[mode].height,
			(m == GT_1BIT) ? "2" :
			(m == GT_4BIT) ? "16" :
			(m == GT_8BIT) ? "256" :
			(m == GT_15BIT) ? "32K" :
			(m == GT_16BIT) ? "64K" :
			(m == GT_24BIT) ? "16M"  :
			(m == GT_32BIT) ? "16M32" : "??");
  
	return buff;
}


int vga_getmodenumber(char *name)
{
	int i, x, y, m;
	char buff[1024];

	if (sscanf(name,"G%dx%dx%s", &x, &y, buff) != 3)
		return -1;

	if (strcmp(buff, "2") == 0) 
		m = GT_1BIT;
	else if (strcmp(buff, "16") == 0)
		m = GT_4BIT;
	else if (strcmp(buff, "256") == 0)
		m = GT_8BIT;
	else if (strcmp(buff, "32K") == 0)
		m = GT_15BIT;
	else if (strcmp(buff, "64K") == 0)
		m = GT_16BIT;
	else if (strcmp(buff, "16M") == 0)
		m = GT_24BIT;
	else if (strcmp(buff, "16M32") == 0)
		m = GT_32BIT;
	else
		return -1;

	for (i = 0; i <= __GLASTMODE;i++) {
		if (ModeInfo[i].width == x &&
		    ModeInfo[i].height == y &&
		    ModeInfo[i].gt == m)
			return i;
	}

	return -1;
}

int vga_lastmodenumber()
{
	return __GLASTMODE;
}


/******************
 *  Memory access *
 ******************/

unsigned char *vga_getgraphmem(void)
{
	return graph_mem;
}

void vga_setpage(int p)
{
	_GSWPRINT_I("GSW: vga_setpage() - %d\n", p);

	if (_ggigsw_pageemu != GSW_HAVEIT) {
		return;
	} else {
		int copysize = GSW_PAGESIZE*(_ggigsw_curpage+1)
			> _ggigsw_fbmem ? _ggigsw_fbmem % GSW_PAGESIZE
			: GSW_PAGESIZE;
		if (_ggigsw_pflip == GSW_HAVEIT) {
			if (_ggigsw_curframe == 0) {
				memcpy(_ggigsw_realfb
				       +(GSW_PAGESIZE*_ggigsw_curpage),
				       _ggigsw_pagefb, copysize);
			} else {
				memcpy(_ggigsw_realfb2
				       +(GSW_PAGESIZE*_ggigsw_curpage),
				       _ggigsw_pagefb, copysize);
			}
			if (p == 0) {
				if (_ggigsw_curframe == 0) {	
					ggiSetDisplayFrame(_ggigsw_visual,
							   0);
					ggiSetReadFrame(_ggigsw_visual, 0);
					ggiSetWriteFrame(_ggigsw_visual, 0);
					_ggigsw_curframe = 1;
				} else {
					ggiSetDisplayFrame(_ggigsw_visual,
							   1);
					ggiSetReadFrame(_ggigsw_visual, 1);
					ggiSetWriteFrame(_ggigsw_visual, 1);
					_ggigsw_curframe = 0;
				}
			}
		} else {
			memcpy(_ggigsw_realfb
			       +(GSW_PAGESIZE*_ggigsw_curpage),
			       _ggigsw_pagefb, copysize);
		}
		_ggigsw_curpage = p;
	}
}


/* Not supported */
void vga_setreadpage(int p)  { }
void vga_setwritepage(int p) { }

void vga_setlogicalwidth(int w)
{
	/* BUGBUG - is set virtualsize. Should be done with modesetting in GGI. */
}

void vga_setdisplaystart(int a)
{
	/* BUGBUG - not supported */
}

void vga_waitretrace(void)
{
	/* BUGBUG - not supported */
}

int vga_claimvideomemory(int m)
{
	if (m > _ggigsw_stride * _ggigsw_mode.virt.y) {
		return -1;
	}
	return 0;
}

void vga_disabledriverreport(void)
{
	/* BUGBUG - not supported */
}

int vga_setmodeX(void)
{
	/* BUGBUG - not supported */
	return -1;
}


int vga_getmonitortype(void)
{
	return MON1024_72;
}

/******************
 * Event handling *
 ******************/

int vga_getmousetype(void)
{
	return MOUSE_MOUSESYSTEMS;
}

void vga_setmousesupport(int s)
{
	_ggigsw_wantmouse = 2;
}

/* Not supported */
void vga_lockvc(void)   { }
void vga_unlockvc(void) { }


int vga_getkey(void)
{
	return ggiKbhit(_ggigsw_visual);
}

/*****************
 * Writing stuff *
 *****************/

int vga_oktowrite(void)
{
	return 1;
}

void vga_copytoplanar256(unsigned char *virtualp, int pitch,
			 int voffset, int vpitch, int w, int h)
{
	/* BUGBUG - not supported */
	return;
}

void vga_copytoplanar16(unsigned char *virtualp, int pitch,
			int voffset, int vpitch, int w, int h)
{
	/* BUGBUG - not supported */
	return;
}

void vga_copytoplane(unsigned char *virtualp, int pitch,
		     int voffset, int vpitch, int w, int h, int plane)
{
	/* BUGBUG - not supported */
	return;
}

int vga_setlinearaddressing(void)
{
	if (_ggigsw_dbuf != NULL) {
		return _ggigsw_stride * _ggigsw_mode.virt.y;
	} else {
		return -1;
	}
}

/* Not supported */
void vga_safety_fork(void (*dummy) (void)) { }


/* Not supported */
void vga_setchipset(int c) { }
void vga_setchipsetandfeatures(int c, int par1, int par2) { }
void vga_gettextfont(void *font) { }
void vga_puttextfont(void *font) { }
void vga_settextmoderegs(void *regs) { }
void vga_gettextmoderegs(void *regs) { }


/*
***************************************************************************
 Acceleration - unsupported
***************************************************************************
*/

int vga_white(void)
{
	ggi_color col = { (1<<16)-1, (1<<16)-1, (1<<16)-1 };

	if (_ggigsw_inggi < 0) return ~0;

	return ggiMapColor(_ggigsw_visual, &col);
}

int vga_setrgbcolor(int r, int g, int b)
{
	ggi_color col;
	ggi_pixel pixval;

	if (_ggigsw_inggi < 1) return -1;

	col.r = r;
	col.g = g;
	col.b = b;

	pixval = ggiMapColor(_ggigsw_visual, &col);
	ggiSetGCForeground(_ggigsw_visual, pixval);

	return pixval;
}

int vga_setegacolor(int c)
{
	/* EGA colors */
	static const unsigned char egared[16] =
	{0,   0,   0,   0, 168, 168, 168, 168, 84,  84,  84,  84, 255, 255, 255, 255};
	static const unsigned char egagreen[16] =
	{0,   0, 168, 168,   0,   0,  84, 168, 84,  84, 255, 255,  84,  84, 255, 255};
	static const unsigned char egablue[16] =
	{0, 168,   0, 168,   0, 168,   0, 168, 84, 255,  84, 255,  84, 255,  84, 255};

	if (_ggigsw_inggi < 0) return -1;

	if (c < 0) c = 0;
	else if (c > 15) c = 15;

	switch (ModeInfo[_ggigsw_inggi].gt) {
	case GT_15BIT:
	case GT_16BIT:
	case GT_24BIT:
	case GT_32BIT:
		return vga_setrgbcolor(egared[c], egagreen[c], egablue[c]);
	default:
		break;
	}

	ggiSetGCForeground(_ggigsw_visual, c);
	return c;
}


/*
***************************************************************************
 Acceleration - unsupported
***************************************************************************
*/

void vga_bitblt(int srcaddr, int destaddr, int w, int h, int pitch)     { }
void vga_imageblt(void *srcaddr, int destaddr, int w, int h, int pitch) { }
void vga_fillblt(int destaddr, int w, int h, int pitch, int c)          { }
void vga_hlinelistblt(int ymin, int n, int *xmin, int *xmax, int pitch, int c)
{ }
void vga_blitwait(void) { }
int vga_ext_set(unsigned what,...)    { return -1; }
int vga_accel(unsigned operation,...) {	return -1; }

    extern int vga_waitevent(int which, fd_set * in, fd_set * out, fd_set * except,
			     struct timeval *timeout);

/* Not supported */
void vga_runinbackground(int stat, ...) { }

int vga_runinbackground_version(void)
{
	/* Tell apps not to use vga_runinbackground() */
	return 0;
}
